/*
Module : SocksSocket.cpp
Purpose: Implementation for a socks socket class
Created: PJN / 16-02-2003
History: PJN / 07-09-2005 1. Fixed a potential buffer overflow issue in the parsing of SOCKS4 connections requests.                    
         PJN / 15-01-2007 1. Fixed a number of memory leaks in the CSocksSocket::Read methods
                          2. Updated code to compile with latest version of the author's CWSocket class

Copyright (c) 2003 - 2017 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/

//////////////// Includes ////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"
#include "SocksSocket.h"


//////////////// Macros //////////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////// Implementation //////////////////////////////////////

bool CSocksSocket::Read(_In_ CSocksConnectionRequest& request, _In_ DWORD dwTimeout, _In_ CWaitableTimer& timer, _In_ HANDLE hStopEvent, _In_ HANDLE hDataEvent)
{
  //Validate our parameters
  ASSERT(m_hSocket != INVALID_SOCKET);

  //The local variables which will receive the data
  BYTE byRawRequest[257]; //257 is max(sizeof(SOCK4_CONNECT_REQUEST), sizeof(SOCKS5_IDENTIFIER_PACKET));

  //retrieve the reponse
  DWORD dwCurrentReadOffset = 0;
  bool bMoreDataToRead = true;
  while (bMoreDataToRead)
  {
    HANDLE hWaitHandles[3];
    hWaitHandles[0] = hDataEvent;
    hWaitHandles[1] = timer;
    hWaitHandles[2] = hStopEvent;

    //Reset the timer
    if (!timer.SetOnceOffRelative(dwTimeout))
    {
      TRACE(_T("CSocksSocket::Read, An error occurred setting the idle timer, Error:%d\n"), GetLastError());
      return false;
    }

    //Wait for something interesting to happen
    DWORD dwWait = WaitForMultipleObjects(3, hWaitHandles, FALSE, INFINITE);
    int nSignaledHandle = dwWait - WAIT_OBJECT_0;

    //Work out what the return value from WFMO means!
    if (nSignaledHandle == 0)
    {
      //receive the data from the socket
      int nBufRemaining = sizeof(byRawRequest) - dwCurrentReadOffset;
      if (nBufRemaining < 0)
        nBufRemaining = 0;

      int nData = 0;
      try
      {
        if (IsReadible(0))
          nData = Receive(byRawRequest + dwCurrentReadOffset, nBufRemaining);
      }
    #ifdef CWSOCKET_MFC_EXTENSIONS
      catch(CWSocketException* pEx)
      {
        TRACE(_T("CSocksSocket::Read, An error occurred reading data from the socket, Error:%d\n"), pEx->m_nError);
        pEx->Delete();
        return false;
      }
    #else
      catch(CWSocketException& e)
      {
        UNREFERENCED_PARAMETER(e);
        TRACE(_T("CSocksSocket::Read, An error occurred reading data from the socket, Error:%d\n"), e.m_nError);
        return false;
      }
    #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

      if (nData == 0)
      {
        TRACE(_T("CSocksSocket::Read, A graceful disconnect has occured reading data from the socket\n"));
        return false;
      }

      //Increment the count of data received
      dwCurrentReadOffset += nData;

      //Try to parse out what we received
      if (dwCurrentReadOffset) 
      {
        SOCK4_CONNECT_REQUEST* pSocks4Request = reinterpret_cast<SOCK4_CONNECT_REQUEST*>(byRawRequest);
        if (pSocks4Request->VN == 4)
        {
          request.m_bSocks5Request = false;

          if (dwCurrentReadOffset >= (sizeof(SOCK4_CONNECT_REQUEST)))
          {
            memcpy(&request.m_Socks4Request, pSocks4Request, sizeof(request.m_Socks4Request));
            request.m_Socks4Request.DSTPORT = ntohs(request.m_Socks4Request.DSTPORT);
          }

          //Have we got all the user ID yet
          int nUserIdRead = dwCurrentReadOffset - sizeof(SOCK4_CONNECT_REQUEST) + sizeof(pSocks4Request->USERID);
          for (int i=0; i<nUserIdRead && bMoreDataToRead; i++)
            bMoreDataToRead = (pSocks4Request->USERID[i] != '\0');
          if (!bMoreDataToRead)
            request.m_sSocks4Ident = pSocks4Request->USERID;
        }
        else if (pSocks4Request->VN == 5)
        {
          SOCKS5_IDENTIFIER_PACKET* pSocks5Request = reinterpret_cast<SOCKS5_IDENTIFIER_PACKET*>(byRawRequest);

          request.m_bSocks5Request = true;
          request.m_Socks5Request.VER = 5;
          if (dwCurrentReadOffset >= 2)
          {
            request.m_Socks5Request.NMETHODS = pSocks5Request->NMETHODS;
            int nMethodsRead = dwCurrentReadOffset - 2;
            bMoreDataToRead = !(nMethodsRead == pSocks5Request->NMETHODS);

            if (!bMoreDataToRead)
              memcpy(request.m_Socks5Request.METHODS, pSocks5Request->METHODS, nMethodsRead);
          }
        }
      }
    }
    else if (nSignaledHandle == 1)
    {
      TRACE(_T("CSocksSocket::Read, Timed out waiting for response from socket\n"));
      return false;
    }
    else
    {
      TRACE(_T("CSocksSocket::Read, Thread being asked to exit\n"));
      return false;
    }
  }

  return true;
}

bool CSocksSocket::Read(_In_ CSocks5RequestDetails& request, _In_ DWORD dwTimeout, _In_ CWaitableTimer& timer, _In_ HANDLE hStopEvent, _In_ HANDLE hDataEvent)
{
  //Validate our parameters
  ASSERT(m_hSocket != INVALID_SOCKET);

  //The local variables which will receive the data
  BYTE byRawRequest[263]; //263 is max(max(sizeof(SOCKS5_IP4_REQUEST_DETAILS), sizeof(SOCKS5_IP6_REQUEST_DETAILS)), sizeof(SOCKS5_HOSTNAME_REQUEST_DETAILS))
  
  //retrieve the reponse
  DWORD dwCurrentReadOffset = 0;
  bool bMoreDataToRead = true;
  while (bMoreDataToRead)
  {
    HANDLE hWaitHandles[3];
    hWaitHandles[0] = hDataEvent;
    hWaitHandles[1] = timer;
    hWaitHandles[2] = hStopEvent;

    //Reset the timer
    if (!timer.SetOnceOffRelative(dwTimeout))
    {
      TRACE(_T("CSocksSocket::Read, An error occurred setting the idle timer, Error:%d\n"), GetLastError());
      return false;
    }

    //Wait for something interesting to happen
    DWORD dwWait = WaitForMultipleObjects(3, hWaitHandles, FALSE, INFINITE);
    int nSignaledHandle = dwWait - WAIT_OBJECT_0;

    //Work out what the return value from WFMO means!
    if (nSignaledHandle == 0)
    {
      //receive the data from the socket
      int nBufRemaining = sizeof(byRawRequest) - dwCurrentReadOffset;
      if (nBufRemaining < 0)
        nBufRemaining = 0;

      int nData = 0;
      try
      {
        if (IsReadible(0))
          nData = Receive(byRawRequest + dwCurrentReadOffset, nBufRemaining);
      }
    #ifdef CWSOCKET_MFC_EXTENSIONS
      catch(CWSocketException* pEx)
      {
        TRACE(_T("CSocksSocket::Read, An error occurred reading data from the socket, Error:%d\n"), pEx->m_nError);
        pEx->Delete();
        return false;
      }
    #else
      catch(CWSocketException& e)
      {
        UNREFERENCED_PARAMETER(e);
        TRACE(_T("CSocksSocket::Read, An error occurred reading data from the socket, Error:%d\n"), e.m_nError);
        return false;
      }
    #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

      //Increment the count of data received
      dwCurrentReadOffset += nData;

      //Try to parse out what we received
      if (dwCurrentReadOffset >= sizeof(SOCKS5_BASE_REQUEST_DETAILS)) 
      {
        SOCKS5_BASE_REQUEST_DETAILS* pBaseRequest = reinterpret_cast<SOCKS5_BASE_REQUEST_DETAILS*>(byRawRequest);
        if (pBaseRequest->ATYP == 1)
        {
          //An IPv4 address type destination
          request.m_Type = CSocks5RequestDetails::IP4;
          bMoreDataToRead = (dwCurrentReadOffset < sizeof(SOCKS5_IP4_REQUEST_DETAILS));
          if (!bMoreDataToRead)
          {
            memcpy(&request.IP_4, byRawRequest, sizeof(SOCKS5_IP4_REQUEST_DETAILS));
            request.IP_4.DSTPORT = ntohs(request.IP_4.DSTPORT);
          }
        }
        else if (pBaseRequest->ATYP == 3)
        {
          request.m_Type = CSocks5RequestDetails::HostName;

          //A host name type destination
          if (dwCurrentReadOffset > sizeof(SOCKS5_BASE_REQUEST_DETAILS))
          {
            SOCKS5_HOSTNAME_REQUEST_DETAILS* pHostnameRequest = reinterpret_cast<SOCKS5_HOSTNAME_REQUEST_DETAILS*>(byRawRequest);
            bMoreDataToRead = (dwCurrentReadOffset < ((sizeof(SOCKS5_HOSTNAME_REQUEST_DETAILS) - 256) + pHostnameRequest->DST_HOST.LENGTH));
            if (!bMoreDataToRead)
            {
              memcpy(&request.HOSTNAME, byRawRequest, sizeof(SOCKS5_HOSTNAME_REQUEST_DETAILS));
              WORD* pPort = reinterpret_cast<WORD*>(&(pHostnameRequest->DST_HOST.HOST[pHostnameRequest->DST_HOST.LENGTH]));
              request.HOSTNAME.DSTPORT = ntohs(*pPort);

              //null terminate the host name
              request.HOSTNAME.DST_HOST.HOST[request.HOSTNAME.DST_HOST.LENGTH] = '\0';
            }
          }
        }
        else if (pBaseRequest->ATYP == 4)
        {
          request.m_Type = CSocks5RequestDetails::IP6;

          //An IPv6 address type destination
          bMoreDataToRead = (dwCurrentReadOffset < sizeof(SOCKS5_IP6_REQUEST_DETAILS));
          if (!bMoreDataToRead)
          {
            memcpy(&request.IP_6, byRawRequest, sizeof(SOCKS5_IP6_REQUEST_DETAILS));
            request.IP_6.DSTPORT = ntohs(request.IP_6.DSTPORT);
          }
        }
        else
        {
          TRACE(_T("CSocksSocket::Read, An invalid destination address type was specified, Address Type:%d\n"), pBaseRequest->ATYP);
          return false;
        }
      }
    }
    else if (nSignaledHandle == 1)
    {
      TRACE(_T("CSocksSocket::Read, Timed out waiting for response from socket\n"));
      return false;
    }
    else
    {
      TRACE(_T("CSocksSocket::Read, Thread being asked to exit\n"));
      return false;
    }
  }

  return true;
}

bool CSocksSocket::Read(_In_ CSocksUsernameAuthenticationDetails& request, _In_ DWORD dwTimeout, _In_ CWaitableTimer& timer, _In_ HANDLE hStopEvent, _In_ HANDLE hDataEvent)
{
  //Validate our parameters
  ASSERT(m_hSocket != INVALID_SOCKET);

  //The local variables which will receive the data
  BYTE byRawRequest[515]; //515 is the maximum size of a user authentication packet
  
  //retrieve the reponse
  int nCurrentReadOffset = 0;
  bool bMoreDataToRead = true;
  while (bMoreDataToRead)
  {
    HANDLE hWaitHandles[3];
    hWaitHandles[0] = hDataEvent;
    hWaitHandles[1] = timer;
    hWaitHandles[2] = hStopEvent;

    //Reset the timer
    if (!timer.SetOnceOffRelative(dwTimeout))
    {
      TRACE(_T("CSocksSocket::Read, An error occurred setting the idle timer, Error:%d\n"), GetLastError());
      return false;
    }

    //Wait for something interesting to happen
    DWORD dwWait = WaitForMultipleObjects(3, hWaitHandles, FALSE, INFINITE);
    int nSignaledHandle = dwWait - WAIT_OBJECT_0;

    //Work out what the return value from WFMO means!
    if (nSignaledHandle == 0)
    {
      //receive the data from the socket
      int nBufRemaining = sizeof(byRawRequest) - nCurrentReadOffset;
      if (nBufRemaining < 0)
        nBufRemaining = 0;

      int nData = 0;
      try
      {
        if (IsReadible(0))
          nData = Receive(byRawRequest + nCurrentReadOffset, nBufRemaining);
      }
    #ifdef CWSOCKET_MFC_EXTENSIONS
      catch(CWSocketException* pEx)
      {
        TRACE(_T("CSocksSocket::Read, An error occurred reading data from the socket, Error:%d\n"), pEx->m_nError);
        pEx->Delete();
        return false;
      }
    #else
      catch(CWSocketException& e)
      {
        UNREFERENCED_PARAMETER(e);
        TRACE(_T("CSocksSocket::Read, An error occurred reading data from the socket, Error:%d\n"), e.m_nError);
        return false;
      }
    #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

      //Increment the count of data received
      nCurrentReadOffset += nData;

      //Try to parse out what we received
      if (nCurrentReadOffset) 
      {
        if (byRawRequest[0] != 1)
        {
          TRACE(_T("CSocksSocket::Read, An invalid negotiation value was found in the response from the socket\n"));
          return false;
        }
        if (nCurrentReadOffset > 1)
        {
          int nUserLength = byRawRequest[1];
          if (nUserLength == 0)
          {
            TRACE(_T("CSocksSocket::Read, An invalid username length value was found in the response from the socket\n"));
            return FALSE;
          }
          if (nCurrentReadOffset >= (nUserLength + 2))
          {
            char szUser[256]; //255 is the maximum size of the username
            memcpy_s(szUser, sizeof(szUser), byRawRequest+2, nUserLength);
            szUser[nUserLength] = '\0'; //Null terminate the data
            request.sUserName = szUser;
            SecureZeroMemory(szUser, sizeof(szUser));
            if (nCurrentReadOffset >= (nUserLength + 3))
            {
              int nPasswordLength = byRawRequest[nUserLength + 2];
              if (nPasswordLength == 0)
              {
                TRACE(_T("CSocksSocket::Read, An invalid password length value was found in the response from the socket\n"));
                return FALSE;
              }
              if (nCurrentReadOffset >= (nUserLength + nPasswordLength + 3))
              {
                char szPassword[256]; //255 is the maximum size of the password
                memcpy_s(szPassword, sizeof(szPassword), byRawRequest+3+nUserLength, nPasswordLength);
                szPassword[nPasswordLength] = '\0'; //Null terminate the data
                request.sPassword = szPassword;
                SecureZeroMemory(szPassword, sizeof(szPassword));

                bMoreDataToRead = false;
              }
            }
          }
        }
      }
    }
    else if (nSignaledHandle == 1)
    {
      TRACE(_T("CSocksSocket::Read, Timed out waiting for response from socket\n"));
      return false;
    }
    else
    {
      TRACE(_T("CSocksSocket::Read, Thread being asked to exit\n"));
      return false;
    }
  }

  return true;
}

void CSocksSocket::SendWithRetry(_In_reads_bytes_(nBufLen) const void* pBuffer, _In_ int nBufLen, _In_ DWORD dwTimeout)
{
  bool bSent = false;

  do
  {
    try
    {
      //Let the socket do the send
      CWSocket::Send(pBuffer, nBufLen);
      bSent = true;
    }
  #ifdef CWSOCKET_MFC_EXTENSIONS
    catch(CWSocketException* pEx)
    {
      //Pull out the exceptions details before we delete it
      int nError = pEx->m_nError;

      //Delete the exception before we go any further
      pEx->Delete();
  #else
    catch(CWSocketException& e)
    {
      //Pull out the exceptions details before we delete it
      int nError = e.m_nError;
  #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

      if (nError == WSAEWOULDBLOCK)
      {
        //If a WSAEWOULDBLOCK error occurred then wait until the socket becomes
        //writtable or the timeout occurs
        int nEvent = WSAEventSelect(*this, m_WSABlockEvent, FD_WRITE);

        //Handle the error
        if (nEvent != 0)
          CWSocket::ThrowWSocketException();

        //Wait for the socket to become writable
        DWORD dwWait = WaitForSingleObject(m_WSABlockEvent, dwTimeout);

        //Cancel the notification
        WSAEventSelect(*this, m_WSABlockEvent, 0);

        //If the socket does not become writable in the timeout period, 
        //then rethrow the exception
        if (dwWait != WAIT_OBJECT_0)
          CWSocket::ThrowWSocketException(WSAEWOULDBLOCK);
      }
      else
      {
        //We caught an exception which we do not know how to handle.
        //In this case just rethrow the exception
        CWSocket::ThrowWSocketException(nError);
      }
    }
  }
  while (!bSent);
}
