/*
Module : SocksSvc.h
Purpose: Defines the interface for an MFC class encapsulation of a Socks server
Created: PJN / 16-02-2003
History: PJN / 03-03-2003 1. Fixed a bug in CSocksSocket::Read(CSocks5RequestDetails&,...) which was causing the port details
                          to be read incorrectly from clients.
         PJN / 07-09-2005 1. Removed the m_bListenBind variable from the CSocksServerSettings as m_bListenBindAddress on its
                          own is sufficient to specify whether binding should occur.
                          2. Removed the m_bConnectBind variable from the CSocksServerSettings as m_bConnectBindAddress on 
                          its own is sufficient to specify whether binding should occur.
         PJN / 15-01-2007 1. Updates to allow the code to clean compile on VC 2005
                          2. Optimized CSocksServerSettings constructor code
                          3. Optimized CSocksServer constructor code
         PJN / 10-02-2017 1. Updated copyright details.
                          2. Replaced BOOL with bool throughout codebase
                          3. Replaced NULL with nullptr throughout codebase
                          4. Updated the code to compile cleanly in VC 2013 - 2015. The code now requires at a minimum VC 2013
                          5. Updated the code to handle IPv6 Socks v5 support
                          6. Added SAL annotations to all the code
                          7. Fixed up logic in CSocksClient::DoSocks5Negotiation to pick an appropriate socks5 client method
                          8. Reworked all CSocksSocket::Read methods to avoid the use of "new".
                          9. Updated all the CSocksSocket::Read methods to handle WSAEWOULDBLOCK errors.
                          10. Updated CSocksClient::Relay method to handle WSAEWOULDBLOCK errors.
                          11. Updated CSocksClient::SendConnectionResponsev5 method to return bound server address and port 
                          number in IPv4 and IPv6 connection responses.

Copyright (c) 2003 - 2017 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////// Macros / Defines //////////////////////////////

#pragma once

#ifndef __SOCKSSVC_H__
#define __SOCKSSVC_H__


/////////////////////////////// Includes //////////////////////////////////////

#include "ThrdPool.h"                //If the code fails to compile because of this missing include then you need to download 
                                     //the CThreadPoolServer classes from my web site and copy the relevant files to the SocksSvc directory
#include "DirectedThreadPoolQueue.h" //If the code fails to compile because of this missing include then you need to download 
                                     //the CThreadPoolServer classes from my web site and copy the relevant files to the SocksSvc directory
#include "IOCPThreadPoolQueue.h"     //If the code fails to compile because of this missing include then you need to download 
                                     //the CThreadPoolServer classes from my web site and copy the relevant files to the SocksSvc directory
#include "SocksClient.h"


/////////////////////////////// Classes ///////////////////////////////////////

//Class which handles directed requests for the socks server
class SOCKSSVC_EXT_CLASS CSocksSvcDirectedThreadPoolQueue : public CDirectedThreadPoolQueue
{
public:
//Constructors / Destructors
  virtual ~CSocksSvcDirectedThreadPoolQueue()
  {
    //Before we are destroyed, deallocate any user requests
    //which may still be lying around in the requests array
    for (std::vector<CThreadPoolRequest>::size_type i=0; i<m_Requests.size(); i++)
    {
      CThreadPoolRequest& request = m_Requests[i];
      if (request.m_dwID == CThreadPoolRequest::THREADPOOL_USER_DEFINED_REQUEST)
      {
        CSocksThreadPoolRequest* pSocksRequest = static_cast<CSocksThreadPoolRequest*>(request.m_pData);
        delete pSocksRequest;
        pSocksRequest = nullptr;
      }
    }
  }
};

//The actual socks server
template <class TCLIENT, class TQUEUE>
class CSocksServer : public ISocksSvcError
{
public:
//Constructors / Destructors
  CSocksServer() : m_ListenStartEvent(TRUE),
                   m_pListenThread(nullptr),
                   m_pSettings(nullptr),
                   m_dwListenerLastError(ERROR_SUCCESS)
  {
  }
  ~CSocksServer()
  {
    Stop();
  }

//Methods
  virtual bool Start(_In_ CSocksServerSettings* pSettings)
  {
    CSingleLock sl(&m_csListenThread, TRUE); //synchronize access to the listening thread

    //Validate our parameters
    ASSERT(m_pListenThread == nullptr); //Trying to start an already started server
    ASSERT(pSettings != nullptr);

    //Start the listener thread
    m_pSettings = pSettings;
    m_dwListenerLastError = ERROR_SUCCESS;
    m_StopEvent.ResetEvent();
    m_ListenStartEvent.ResetEvent();
    m_pListenThread = AfxBeginThread(ListenSocketFunction, this, THREAD_PRIORITY_NORMAL, 0, CREATE_SUSPENDED);
    if (m_pListenThread == nullptr)
    {
      //Report the error
      CString sError;
      sError.Format(_T("CSocksServer::Start, Failed to call to create listener thread, please check settings, Error:%u"), GetLastError());
      OnError(sError);
      return false;
    }
    m_pListenThread->m_bAutoDelete = FALSE;
    m_pListenThread->ResumeThread();

    //Wait until the thread has completely started and return the success indicator
    WaitForSingleObject(m_ListenStartEvent, INFINITE);
    SetLastError(m_dwListenerLastError);
    return (m_dwListenerLastError == ERROR_SUCCESS);
  }

  virtual bool Stop()
  {
    //Signal the listener thread to stop and wait for it to do so, Note that we do 
    //this before we enter the critical section because otherwise we could end
    //up in a deadlock situation between this function and Wait
    m_StopEvent.SetEvent();

    CSingleLock sl(&m_csListenThread, TRUE); //synchronize access to the listening thread

    //If the listener thread is running, then stop it
    if (m_pListenThread != nullptr)
    {
      WaitForSingleObject(m_pListenThread->m_hThread, INFINITE);
      delete m_pListenThread;
      m_pListenThread = nullptr;
    }

    return true;
  }

  virtual bool Wait()
  {
    CSingleLock sl(&m_csListenThread, TRUE); //synchronize access to the listening thread

    //If the listener thread is running, then just wait for it to exit
    if (m_pListenThread != nullptr)
    {
      WaitForSingleObject(m_pListenThread->m_hThread, INFINITE);
      delete m_pListenThread;
      m_pListenThread = nullptr;
      return true;
    }
    else
    {
      //Report the error
      CString sError;
      sError.Format(_T("CSocksServer::Wait, Socks server is not running, so no need to wait"));
      OnError(sError);
      return false;
    }
  }

  virtual void OnError(_In_ const CString& sError)
  {
  #ifdef _DEBUG
    //The default is to just TRACE the details
    TRACE(_T("%s\n"), sError.GetString());
  #else
    UNREFERENCED_PARAMETER(sError);
  #endif //#ifdef _DEBUG
  }

  CThreadPoolServer<TCLIENT, TQUEUE>& GetThreadPool() 
  { 
    return m_ThreadPool; 
  }

protected:
//Methods
  static UINT ListenSocketFunction(LPVOID pParam)
  {
    //Get back the "this" pointer from the pParam parameter
    CSocksServer* pServer = static_cast<CSocksServer*>(pParam);
    ASSERT(pServer != nullptr);

    //Call the run method of the CSocksServer instance
    pServer->ListenSocketFunction();

    //Return the thread exit code
    return 0;
  }

  virtual void ListenSocketFunction()
  {
    //Validate our parameters
    AFXASSUME(m_pSettings != nullptr);

    //Setup the thread pool 
    if (!m_ThreadPool.Start(m_pSettings->m_nThreadPoolSize, m_pSettings->m_nThreadPoolSize, TRUE))
    {
      //Report the error
      m_dwListenerLastError = GetLastError();
      CString sError;
      sError.Format(_T("CSocksServer::ListenSocketFunction, Failed to create thread pool, Error:%u"), m_dwListenerLastError);
      OnError(sError);
      m_ListenStartEvent.SetEvent();
      return;
    }

    //Setup the pointers in the thread pool instances and resume the threads
    for (int i=0; i<m_pSettings->m_nThreadPoolSize; i++)
    {
      CSocksClient* pClient = static_cast<CSocksClient*>(m_ThreadPool.GetAtClient(i));
      AFXASSUME(pClient != nullptr);
      pClient->m_pError = this;
      pClient->m_pSettings = m_pSettings;
      HANDLE hWorkerThread = pClient->GetThreadHandle();
      AFXASSUME(hWorkerThread != nullptr);
      if (ResumeThread(hWorkerThread) == static_cast<DWORD>(-1))
      {
        //Report the error
        m_dwListenerLastError = GetLastError();
        CString sError;
        sError.Format(_T("CSocksServer::ListenSocketFunction, Failed to resume worker thread at index %d in thread pool, Error:%u"), i, m_dwListenerLastError);
        OnError(sError);

        m_ListenStartEvent.SetEvent();
        return;
      }
    }

    //Setup the recycling of threads
    if (!m_ThreadPool.SetMaxThreadClientLifetime(m_pSettings->m_bEnableThreadLifetime, m_pSettings->m_dwThreadLifetime))
    {
      //Report the error
      m_dwListenerLastError = GetLastError();
      CString sError;
      sError.Format(_T("CSocksServer::ListenSocketFunction, Failed in call to m_ThreadPool.SetMaxThreadClientLifetime, Error:%u"), m_dwListenerLastError);
      OnError(sError);

      m_ListenStartEvent.SetEvent();
      return;
    }

    //Create and bind the server socket
    CSocksSocket serverSocket;
    try
    {
    #ifdef CWSOCKET_MFC_EXTENSIONS
      serverSocket.SetBindAddress(m_pSettings->m_sListenBindAddress);
    #else
      serverSocket.SetBindAddress(CWSocket::String(m_pSettings->m_sListenBindAddress));
    #endif //#ifdef CWSOCKET_MFC_EXTENSIONS
      serverSocket.CreateAndBind(m_pSettings->m_nPort, SOCK_STREAM, m_pSettings->m_bIPv6 ? AF_INET6 : AF_INET);
    }
  #ifdef CWSOCKET_MFC_EXTENSIONS
    catch(CWSocketException* pEx)
    {
      m_dwListenerLastError = pEx->m_nError;

      //Report the error
      CString sError;
      sError.Format(_T("CSocksServer::ListenSocketFunction, Failed to bind server socket, Error:%d"), pEx->m_nError);
      OnError(sError);

      m_ListenStartEvent.SetEvent();
      pEx->Delete();  
      return;
    }
  #else
    catch(CWSocketException& e)
    {
      m_dwListenerLastError = e.m_nError;

      //Report the error
      CString sError;
      sError.Format(_T("CSocksServer::ListenSocketFunction, Failed to bind server socket,Error:%d"), e.m_nError);
      OnError(sError);

      m_ListenStartEvent.SetEvent();
      return;
    }
  #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

    //Put the server socket in a listening state
    try
    {
      serverSocket.Listen();
    }
  #ifdef CWSOCKET_MFC_EXTENSIONS
    catch(CWSocketException* pEx)
    {
      m_dwListenerLastError = pEx->m_nError;

      //Report the error
      CString sError;
      sError.Format(_T("CSocksServer::ListenSocketFunction, Failed to listen on server socket, Error:%d"), pEx->m_nError);
      OnError(sError);

      m_ListenStartEvent.SetEvent();
      pEx->Delete();
      return;
    }
  #else
    catch(CWSocketException& e)
    {
      m_dwListenerLastError = e.m_nError;

      //Report the error
      CString sError;
      sError.Format(_T("CSocksServer::ListenSocketFunction, Failed to listen on server socket, Error:%d"), e.m_nError);
      OnError(sError);

      m_ListenStartEvent.SetEvent();
      return;
    }
  #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

    //Use a Win32 event notification on the server socket
    int nError = WSAEventSelect(serverSocket, m_SocketEvent, FD_ACCEPT);
    if (nError == SOCKET_ERROR)
    {
      m_dwListenerLastError = GetLastError();

      //Report the error
      CString sError;
      sError.Format(_T("CSocksServer::ListenSocketFunction, Failed in call to WSAEventSelect, Error:%u"), m_dwListenerLastError);
      OnError(sError);

      m_ListenStartEvent.SetEvent();
      return;
    }

    //We're now ready for accepting client connections, inform the main thread that everthing is ok
    m_dwListenerLastError = ERROR_SUCCESS;
    m_ListenStartEvent.SetEvent();

    //Create the wait handles array (used for waiting for an event)
    HANDLE Handles[2];
    Handles[0] = m_StopEvent;
    Handles[1] = m_SocketEvent;

    //Wait for any incoming connections and the signal to exit the thread m_ListenStopEvent
    bool bWantStop = false;
    CThreadPoolQueue* pQueue = m_ThreadPool.GetQueue();
    ASSERT(pQueue);
    while (!bWantStop)
    {
      DWORD dwWait = WaitForMultipleObjects(2, Handles, FALSE, INFINITE);
      int nSignaledHandle = dwWait - WAIT_OBJECT_0;
      if (nSignaledHandle == 0) //It was the stop request
        bWantStop = true;
      else if (nSignaledHandle == 1) //It was a socket event
      {
        CSocksThreadPoolRequest* pRequest = new CSocksThreadPoolRequest;  
        try
        {
          //Accept the client connection
          int nClientAddressLength = sizeof(pRequest->m_ClientAddress);
          serverSocket.Accept(pRequest->m_ClientSocket, reinterpret_cast<SOCKADDR*>(&pRequest->m_ClientAddress), &nClientAddressLength);

          //Let a thread pool client class instance handle the work
          CThreadPoolRequest poolRequest;
          poolRequest.m_dwID = CThreadPoolRequest::THREADPOOL_USER_DEFINED_REQUEST;
          poolRequest.m_pData = pRequest;  
          if (!pQueue->PostRequest(poolRequest))
          {
            OnError(_T("CSocksServer::ListenSocketFunction, Failed in post request to thread pool"));
            delete pRequest;
          }
        }
      #ifdef CWSOCKET_MFC_EXTENSIONS
        catch(CWSocketException* pEx)
        {
          //Report the error
          CString sError;
          sError.Format(_T("CSocksServer::ListenSocketFunction, An error occurred accepting a client connection, Error:%d"), pEx->m_nError);
          OnError(sError);

          pEx->Delete();
        }
      #else
        catch(CWSocketException& e)
        {
          //Report the error
          CString sError;
          sError.Format(_T("CSocksServer::ListenSocketFunction, An error occurred accepting a client connection, Error:%d"), e.m_nError);
          OnError(sError);
        }
      #endif //#ifdef CWSOCKET_MFC_EXTENSIONS
      }
    }

    //Bring down the thread pool
    m_ThreadPool.Stop();
  }

  //Member variables
  CWinThread*                        m_pListenThread;
  CCriticalSection                   m_csListenThread;
  CEvent                             m_ListenStartEvent;
  CEvent                             m_StopEvent;
  CEvent                             m_SocketEvent;
  DWORD                              m_dwListenerLastError;
  CSocksServerSettings*              m_pSettings;
  CThreadPoolServer<TCLIENT, TQUEUE> m_ThreadPool;
};

#endif //#ifndef __SOCKSSVC_H__
